/*
 * Linux input event mouse driver for XFree86
 *
 * (C) 2005 by Roman Hodek <roman@hodek.net>
 *
 * Some code was copied from the original XFree86 mouse driver
 * (xc/programs/Xserver/hw/xfree86/input/mouse/mouse.c) and also a little bit
 * from the acecad tablet driver as example of an external input driver.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of copyright holders not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The copyright holders
 * make no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#ifdef LINUX_INPUT
#include <asm/types.h>
#include <linux/input.h>
#ifdef BUS_PCI
#undef BUS_PCI
#endif
#ifdef BUS_ISA
#undef BUS_ISA
#endif
#endif

#include <misc.h>
#include <xf86.h>
#define NEED_XF86_TYPES
#include <xf86_ansic.h>
#include <xf86_OSproc.h>
#include <xisb.h>
#include <xf86Xinput.h>
#include <exevents.h>
#include <xf86Module.h>
#include <xf86OSmouse.h>
#include <mipointer.h>

#include "leventPriv.h"
#include "levent_codes.h"


InputDriverRec levent_driver = {
	1,
	"levent",
	NULL,
	LeventPreInit,
	NULL,
	NULL,
	0
};

static XF86ModuleVersionInfo VersionRec =
{
	"levent",
	MODULEVENDORSTRING,
	MODINFOSTRING1,
	MODINFOSTRING2,
	XF86_VERSION_CURRENT,
	1, 0, 0,
	ABI_CLASS_XINPUT,
	ABI_XINPUT_VERSION,
	MOD_CLASS_XINPUT,
	{0, 0, 0, 0}
};

XF86ModuleData leventModuleData = {
    &VersionRec,
    SetupProc,
    TearDownProc
};

typedef enum {
    OPTION_ALWAYS_CORE,
    OPTION_SEND_CORE_EVENTS,
    OPTION_CORE_POINTER,
    OPTION_HISTORY_SIZE,
    OPTION_EVENTDEV_PREFIX,
    OPTION_EVENTDEV_NAME,
    OPTION_EVENTDEV_UNIQ,
    OPTION_SHOW_DEVINFO,
    OPTION_SHOW_EVENTS,
    MAPPING_OPTION_NAMES
} LeventOpts;

static const OptionInfoRec leventOptions[] = {
    { OPTION_ALWAYS_CORE,	"AlwaysCore",	  OPTV_BOOLEAN,	{0}, FALSE },
    { OPTION_SEND_CORE_EVENTS,	"SendCoreEvents", OPTV_BOOLEAN,	{0}, FALSE },
    { OPTION_CORE_POINTER,	"CorePointer",	  OPTV_BOOLEAN,	{0}, FALSE },
    { OPTION_HISTORY_SIZE,	"HistorySize",	  OPTV_INTEGER,	{0}, FALSE },
    { OPTION_EVENTDEV_PREFIX,	"EventDevPrefix", OPTV_STRING,	{0}, FALSE },
    { OPTION_EVENTDEV_NAME,	"EventDevName",   OPTV_STRING,	{0}, FALSE },
    { OPTION_EVENTDEV_UNIQ,	"EventDevUniq",   OPTV_STRING,	{0}, FALSE },
    { OPTION_SHOW_DEVINFO,	"ShowDeviceInfo", OPTV_BOOLEAN,	{0}, FALSE },
    { OPTION_SHOW_EVENTS,	"ShowEvents",     OPTV_BOOLEAN,	{0}, FALSE },
    MAPPING_OPTIONS
    { -1,			NULL,		  OPTV_NONE,	{0}, FALSE }
};

typedef enum {
    MAPTYPE_NONE,
    MAPTYPE_X,
    MAPTYPE_Y,
    MAPTYPE_BUTTON,
    MAPTYPE_AXIS,
} MapType;

#define MAKE_MAP(type,p0,p1) (((type) << 16) | (((p1)&0xff)<<8) | ((p0)&0xff))
#define MAP_TYPE(map)        ((map) >> 16)
#define MAP_P1(map)          (((map) >> 8) & 0xff)
#define MAP_P0(map)          ((map) & 0xff)

struct defmap {
    unsigned type, code;
    unsigned maptype, map0, map1;
} default_mappings[] = {
    { EV_KEY, BTN_0,		MAPTYPE_BUTTON, 0, 0 },
    { EV_KEY, BTN_LEFT,		MAPTYPE_BUTTON, 0, 0 },
    { EV_KEY, BTN_TOUCH,	MAPTYPE_BUTTON, 0, 0 },
    { EV_KEY, BTN_FORWARD,	MAPTYPE_BUTTON, 0, 0 },
    { EV_KEY, BTN_1,		MAPTYPE_BUTTON, 2, 0 },
    { EV_KEY, BTN_RIGHT,	MAPTYPE_BUTTON, 2, 0 },
    { EV_KEY, BTN_STYLUS,	MAPTYPE_BUTTON, 2, 0 },
    { EV_KEY, BTN_2,		MAPTYPE_BUTTON, 1, 0 },
    { EV_KEY, BTN_MIDDLE,	MAPTYPE_BUTTON, 1, 0 },
    { EV_KEY, BTN_STYLUS2,	MAPTYPE_BUTTON, 1, 0 },
    { EV_KEY, BTN_3,		MAPTYPE_BUTTON, 5, 0 },
    { EV_KEY, BTN_BACK,		MAPTYPE_BUTTON, 5, 0 },
    { EV_KEY, BTN_SIDE,		MAPTYPE_BUTTON, 5, 0 },
    { EV_KEY, BTN_4,		MAPTYPE_BUTTON, 6, 0 },
    { EV_KEY, BTN_EXTRA,	MAPTYPE_BUTTON, 6, 0 },
    { EV_REL, REL_X,		MAPTYPE_X, 0, 0 },
    { EV_REL, REL_Y,		MAPTYPE_Y, 0, 0 },
    { EV_REL, REL_WHEEL,	MAPTYPE_AXIS, 4, 3 },
    { 0, 0, MAPTYPE_NONE, 0, 0 }
};


#ifdef XFree86LOADER

static const OptionInfoRec *
LeventAvailableOptions(void *unused)
{
    return (leventOptions);
}

ModuleInfoRec LeventInfo = {
    1,
    "LEVENT",
    NULL,
    0,
    LeventAvailableOptions
};

static pointer
SetupProc(pointer module,
	  pointer options,
	  int *errmaj,
	  int *errmin)
{
    static Bool Initialised = FALSE;

    if (!Initialised) {
	Initialised = TRUE;
#ifndef REMOVE_LOADER_CHECK_MODULE_INFO
	if (xf86LoaderCheckSymbol("xf86AddModuleInfo"))
#endif
	xf86AddModuleInfo(&LeventInfo, module);
    }

    xf86AddInputDriver(&levent_driver, module, 0);
    return module;
}

static void
TearDownProc(pointer p)
{
}

#endif


static LeventEventType *LeventFindType(int value)
{
    LeventEventType *p;

    for(p = levent_types; p->name; p++) {
	if (p->value == value)
	    return p;
    }
    return NULL;
}

static LeventEventCode *LeventFindCode(LeventEventCode *table, int value)
{
    LeventEventCode *p;

    for(p = table; p->name; p++) {
	if (p->value == value)
	    return p;
    }
    return NULL;
}


#define BITS_PER_LONG (sizeof(long) * 8)
#define NBITS(x) ((((x)-1)/BITS_PER_LONG)+1)
#define test_bit(bit, array)	((array[LONG(bit)] >> OFF(bit)) & 1)
#define OFF(x)  ((x)%BITS_PER_LONG)
#define LONG(x) ((x)/BITS_PER_LONG)

static void
LeventShowDeviceInfo(InputInfoPtr pInfo, int fd)
{
    MouseDevPtr pMse = pInfo->private;
    mousePrivPtr mPriv = (mousePrivPtr)pMse->mousePriv;
    unsigned long bit[EV_MAX][NBITS(KEY_MAX)];
    int i, j;
    int abs[5];
    char msg[256];
    LeventEventType *ev_type;
    LeventEventCode *ev_code;

    memset(bit, 0, sizeof(bit));
    ioctl(fd, EVIOCGBIT(0, EV_MAX), bit[0]);

    for(i = 0; i < EV_MAX; i++) {
	if (test_bit(i, bit[0])) {
	    ev_type = LeventFindType(i);
	    xf86Msg(X_INFO, "%s can send the following type %s events:\n",
		    mPriv->found_device, ev_type ? ev_type->name : "unknown");
	    if (!ev_type || !ev_type->code_table)
		continue;
	    ioctl(fd, EVIOCGBIT(i, KEY_MAX), bit[i]);
	    for(j = 0; j < KEY_MAX; j++) {
		if (test_bit(j, bit[i])) {
		    ev_code = LeventFindCode(ev_type->code_table, j);
		    sprintf(msg, "%s can send ", mPriv->found_device);
		    if (ev_code)
			strcat(msg, ev_code->name);
		    else
			sprintf(msg+strlen(msg), "unknown %s_%u",
				ev_type->name, j);
		    strcat(msg, " events, ");
		    if (ev_code && ev_code->mapping) {
			switch(MAP_TYPE(ev_code->mapping)) {
			  case MAPTYPE_X:
			    strcat(msg, "mapped to X axis\n");
			    break;
			  case MAPTYPE_Y:
			    strcat(msg, "mapped to Y axis\n");
			    break;
			  case MAPTYPE_BUTTON:
			    sprintf(msg+strlen(msg), "mapped to X11 button %u\n",
				    MAP_P0(ev_code->mapping)+1);
			    break;
			  case MAPTYPE_AXIS:
			    sprintf(msg+strlen(msg), "mapped to X11 buttons %u and %u\n",
				    MAP_P0(ev_code->mapping)+1,
				    MAP_P1(ev_code->mapping)+1);
			    break;
			}
		    }
		    else
			strcat(msg, "not mapped\n");
		    xf86Msg(X_INFO, msg);
		    
		    if (i == EV_ABS) {
			ioctl(fd, EVIOCGABS(j), abs);
			xf86Msg(X_INFO, "%s max for %s is %d\n",
				mPriv->found_device,
				ev_code ? ev_code->name : "??",
				abs[2]);
		    }
		}
	    }
	}
    }
}

static int
LeventOpenDevice(InputInfoPtr pInfo)
{
    MouseDevPtr pMse = pInfo->private;
    mousePrivPtr mPriv = (mousePrivPtr)pMse->mousePriv;
    DIR *dir;
    XF86DIRENT *de;
    stat_t st;
    char path[PATH_MAX];
    char *fname;
    int prefix_len;
    int fd = -1;
    int version;
    int version_reported = 0;
    char idev_name[128];
    char idev_uniq[128];

    if (mPriv->found_device) {
	if ((fd = open(mPriv->found_device, O_RDWR)) < 0) {
	    xf86Msg(X_WARNING, "%s: %s: open: %s\n",
		    pInfo->name, mPriv->found_device, strerror(errno));
	}
	return fd;
    }
    
    if (!(dir = opendir(mPriv->dev_dir))) {
	xf86Msg(X_ERROR, "%s: cannot open directory %s: %s\n",
		pInfo->name, mPriv->dev_dir, strerror(errno));
	return -1;
    }
    strcpy(path, mPriv->dev_dir);
    strcat(path, "/");
    fname = path+strlen(path);
    prefix_len = strlen(mPriv->dev_prefix);
    
    while((de = readdir(dir))) {
	/* skip if name doesn't match prefix */
	if (prefix_len > 0 &&
	    strncmp(de->d_name, mPriv->dev_prefix, prefix_len) != 0)
	    continue;

	strcpy(fname, de->d_name);
	if (stat(path, &st)) {
	    xf86Msg(X_WARNING, "%s: cannot stat %s\n",
		    pInfo->name, path);
	    continue;
	}
	/* skip if file isn't a char device */
	if (st.st_rdev == 0)
	    continue;

	/*
	xf86Msg(X_INFO, "%s: checking device %s\n",
		pInfo->name, path);
	*/
	if ((fd = open(path, O_RDWR)) < 0) {
	    xf86Msg(X_WARNING, "%s: %s: open: %s\n",
		    pInfo->name, path, strerror(errno));
	    continue;
	}
	if (ioctl(fd, EVIOCGVERSION, &version) < 0) {
	    xf86Msg(X_WARNING, "%s: %s: EVIOCGVERSION: %s\n",
		    pInfo->name, path, strerror(errno));
	    continue;
	}
	if (!version_reported) {
	    xf86Msg(X_INFO,"Kernel Input driver version is %d.%d.%d\n",
		    version >> 16, (version >> 8) & 0xff, version & 0xff);
	    version_reported = 1;
	}

	if (ioctl(fd, EVIOCGNAME(sizeof(idev_name)), idev_name) < 0) {
	    if (errno != ENOENT) {
		xf86Msg(X_WARNING, "%s: %s: EVIOCGNAME: %s\n",
			pInfo->name, path, strerror(errno));
		close(fd);
		continue;
	    }
	    else
		idev_name[0] = '\0';
	}
	if (ioctl(fd, EVIOCGUNIQ(sizeof(idev_uniq)), idev_uniq) < 0) {
	    if (errno != ENOENT) {
		xf86Msg(X_WARNING, "%s: %s: EVIOCGUNIQ: %s\n",
			pInfo->name, path, strerror(errno));
		close(fd);
		continue;
	    }
	    else
		idev_uniq[0] = '\0';
	}
	xf86Msg(X_INFO, "%s: device %s: name='%s' uniq='%s'\n",
		pInfo->name, path, idev_name, idev_uniq);

	/* check that name and uniq match, if they were defined */
	if (mPriv->dev_name && strlen(mPriv->dev_name) &&
	    strcmp(mPriv->dev_name, idev_name) != 0) {
	    xf86Msg(X_INFO, "%s: device %s: name mismatch\n",
		    pInfo->name, path);
	    close(fd);
	    continue;
	}
	if (mPriv->dev_uniq && strlen(mPriv->dev_uniq) &&
	    strcmp(mPriv->dev_uniq, idev_uniq) != 0) {
	    xf86Msg(X_INFO, "%s: device %s: serial mismatch\n",
		    pInfo->name, path);
	    close(fd);
	    continue;
	}

	mPriv->found_device = xstrdup(path);
	xf86Msg(X_INFO, "%s: using device %s\n",
		pInfo->name, mPriv->found_device);
	if (mPriv->show_device_info)
	    LeventShowDeviceInfo(pInfo, fd);
	break;
    }
    closedir(dir);

    if (mPriv->found_device)
	return fd;
    return -1;
}

static void
LeventCommonOptions(InputInfoPtr pInfo)
{
    MouseDevPtr pMse = pInfo->private;
    mousePrivPtr mPriv = (mousePrivPtr)pMse->mousePriv;
    char *p, *s;
    char dirname[PATH_MAX], prefix[PATH_MAX];
    LeventEventType *ev_type;
    LeventEventCode *ev_code;
    struct defmap *defm;

    s = xf86SetStrOption(pInfo->options, "EventDevPrefix", "/dev/input/event");
    /* split EventDevPrefix into directory and name prefix */
    strcpy(dirname, s);
    prefix[0] = '\0';
    if ((p = strrchr(dirname, '/'))) {
	strcpy(prefix, p+1);
	*p = '\0';
    }
    mPriv->dev_dir = xstrdup(dirname);
    mPriv->dev_prefix = xstrdup(prefix);
    xf86Msg(X_CONFIG, "%s: event device directory '%s', prefix '%s'\n",
	    pInfo->name, mPriv->dev_dir, mPriv->dev_prefix);
    
    s = xf86SetStrOption(pInfo->options, "EventDevName", NULL);
    mPriv->dev_name = s ? xstrdup(s) : NULL;
    s = xf86SetStrOption(pInfo->options, "EventDevUniq", NULL);
    mPriv->dev_uniq = s ? xstrdup(s) : NULL;
    xf86Msg(X_CONFIG, "%s: event device required name '%s' and uniq '%s'\n",
	    pInfo->name,
	    mPriv->dev_name ? mPriv->dev_name : "(none)",
	    mPriv->dev_uniq ? mPriv->dev_uniq : "(none)");

    mPriv->show_device_info =
	xf86SetBoolOption(pInfo->options, "ShowDeviceInfo", FALSE);
    mPriv->show_events =
	xf86SetBoolOption(pInfo->options, "ShowEvents", FALSE);

    /* establish default mappings */
    for(defm = default_mappings; defm->maptype != MAPTYPE_NONE; ++defm) {
	if (!(ev_type = LeventFindType(defm->type)))
	    continue;
	if (!ev_type->code_table)
	    continue;
	if (!(ev_code = LeventFindCode(ev_type->code_table, defm->code)))
	    continue;
	ev_code->mapping = MAKE_MAP(defm->maptype, defm->map0, defm->map1);
    }

    /* now parse the users mapping */
    for(ev_type = levent_types; ev_type->name; ev_type++) {
	if (!ev_type->code_table)
	    continue;
	for(ev_code = ev_type->code_table; ev_code->name; ev_code++) {
	    char option_name[64];
	    int mapping = 0;
	    
	    strcpy(option_name, "Map_");
	    strcat(option_name, ev_code->name);
	    if (!(s = xf86SetStrOption(pInfo->options, option_name, NULL)))
		continue;

	    if (strcmp(s, "-") == 0) {
		ev_code->mapping = 0;
	    }
	    else if (strcasecmp(s, "x") == 0) {
		if (ev_type->value != EV_REL || ev_type->value != EV_ABS) {
		    xf86Msg(X_ERROR, "%s: invalid mapping for event '%s': "
			    "only REL and ABS type events can be mapped "
			    "to an axis\n",
			    pInfo->name, ev_code->name);
		    continue;
		}
		mapping = MAKE_MAP(MAPTYPE_X, 0, 0);
	    }
	    else if (strcasecmp(s, "y") == 0) {
		if (ev_type->value != EV_REL || ev_type->value != EV_ABS) {
		    xf86Msg(X_ERROR, "%s: invalid mapping for event '%s': "
			    "only REL and ABS type events can be mapped "
			    "to an axis\n",
			    pInfo->name, ev_code->name);
		    continue;
		}
		mapping = MAKE_MAP(MAPTYPE_Y, 0, 0);
	    }
	    else if (strncasecmp(s, "b", 1) == 0) {
		int b0, b1;

		if (ev_type->value == EV_KEY) {
		    if (sscanf(s+1, "%u", &b0) != 1) {
			xf86Msg(X_ERROR, "%s: syntax error in button mapping "
				"'%s' for event '%s': need one number\n",
				pInfo->name, s, ev_code->name);
			continue;
		    }
		    if (b0 == 0) {
			xf86Msg(X_ERROR, "%s: bad mapping for event '%s': "
				"button  0 is invalid\n",
				pInfo->name, ev_code->name);
			continue;
		    }
		    if (b0 > MSE_MAXBUTTONS) {
			xf86Msg(X_ERROR, "%s: bad mapping for event '%s': "
				"button number %u too large\n",
				pInfo->name, ev_code->name, b0);
			continue;
		    }
		    mapping = MAKE_MAP(MAPTYPE_BUTTON, b0-1, 0);
		}
		else {
		    if (sscanf(s+1, "%u,%u", &b0, &b1) != 2) {
			xf86Msg(X_ERROR, "%s: syntax error in button mapping "
				"'%s' for event '%s': need two numbers\n",
				pInfo->name, s, ev_code->name);
			continue;
		    }
		    if (b0 == 0) {
			xf86Msg(X_ERROR, "%s: bad mapping for event '%s': "
				"button  0 is invalid\n",
				pInfo->name, ev_code->name);
			continue;
		    }
		    if (b0 > MSE_MAXBUTTONS) {
			xf86Msg(X_ERROR, "%s: bad mapping for event '%s': "
				"button number %u too large\n",
				pInfo->name, ev_code->name, b0);
			continue;
		    }
		    if (b1 == 0) {
			xf86Msg(X_ERROR, "%s: bad mapping for event '%s': "
				"button  0 is invalid\n",
				pInfo->name, ev_code->name);
			continue;
		    }
		    if (b1 > MSE_MAXBUTTONS) {
			xf86Msg(X_ERROR, "%s: bad mapping for event '%s': "
				"button number %u too large\n",
				pInfo->name, ev_code->name, b1);
			continue;
		    }
		    mapping = MAKE_MAP(MAPTYPE_AXIS, b0-1, b1-1);
		}
	    }
	    ev_code->mapping = mapping;
	}
    }

    /* calculate total number of buttons */
    pMse->buttons = 0;
    for(ev_type = levent_types; ev_type->name; ev_type++) {
	if (!ev_type->code_table)
	    continue;
	for(ev_code = ev_type->code_table; ev_code->name; ev_code++) {
	    switch(MAP_TYPE(ev_code->mapping)) {
	      case MAPTYPE_BUTTON:
		if (MAP_P0(ev_code->mapping)+1 > pMse->buttons)
		    pMse->buttons = MAP_P0(ev_code->mapping)+1;
		break;
	      case MAPTYPE_AXIS:
		if (MAP_P0(ev_code->mapping)+1 > pMse->buttons)
		    pMse->buttons = MAP_P0(ev_code->mapping)+1;
		if (MAP_P1(ev_code->mapping)+1 > pMse->buttons)
		    pMse->buttons = MAP_P1(ev_code->mapping)+1;
		break;
	    }
	}
    }
    xf86Msg(X_INFO, "%s: total number of buttons: %d\n",
	    pInfo->name, pMse->buttons);
    
}

static InputInfoPtr
LeventPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
{
    InputInfoPtr pInfo;
    MouseDevPtr pMse;
    mousePrivPtr mPriv;
    int fd;
    
    if (!(pInfo = xf86AllocateInput(drv, 0)))
	return NULL;

    /* Initialise the InputInfoRec. */
    pInfo->name = dev->identifier;
    pInfo->type_name = XI_MOUSE;
    pInfo->flags = XI86_POINTER_CAPABLE | XI86_SEND_DRAG_EVENTS;
    pInfo->device_control = LeventProc;
    pInfo->read_input = LeventReadInput;
    pInfo->motion_history_proc = xf86GetMotionEvents;
    pInfo->history_size = 0;
    pInfo->control_proc = NULL;
    pInfo->close_proc = NULL;
    pInfo->switch_mode = NULL;
    pInfo->conversion_proc = LeventConvert;
    pInfo->reverse_conversion_proc = NULL;
    pInfo->fd = -1;
    pInfo->dev = NULL;
    pInfo->private_flags = 0;
    pInfo->always_core_feedback = 0;
    pInfo->conf_idev = dev;

    /* Allocate the MouseDevRec and initialise it. */
    if (!(pMse = xcalloc(sizeof(MouseDevRec), 1)))
	return pInfo;
    pInfo->private = pMse;
    pMse->Ctrl = LeventCtrl;
    pMse->PostEvent = NULL;
    pMse->CommonOptions = LeventCommonOptions;
    
    if (!(mPriv = (pointer) xcalloc(sizeof(mousePrivRec), 1)))
	return pInfo;
    mPriv->found_device = NULL;
    mPriv->count = 0;

    /* Collect the options, and process the common options. */
    xf86CollectInputOptions(pInfo, NULL, NULL);
    xf86ProcessCommonOptions(pInfo, pInfo->options);
    pInfo->fd = -1;

    pMse->mousePriv = mPriv;
    pMse->CommonOptions(pInfo);

    if ((fd = LeventOpenDevice(pInfo)) < 0) {
	xf86Msg(X_ERROR, "%s: no matching event device found\n",
		pInfo->name);
	return pInfo;
    }
    close(fd);
    xf86Msg(X_CONFIG, "%s: Linux Event Device Mouse driver, using device %s\n",
	    pInfo->name, mPriv->found_device);
    
    pInfo->flags |= XI86_CONFIGURED;
    return pInfo;
}

static void
LeventReadInput(InputInfoPtr pInfo)
{
    MouseDevPtr pMse = pInfo->private;
    mousePrivPtr mPriv = (mousePrivPtr)pMse->mousePriv;
    int button, down = 0, dx, dy;
    int pBufP;
    int c;
    unsigned char *pBuf, u;
    struct input_event ev;
    LeventEventType *ev_type;
    LeventEventCode *ev_code;

    pBufP = mPriv->count;
    pBuf = mPriv->data;

    /*
     * Set blocking to -1 on the first call because we know there is data to
     * read. Xisb automatically clears it after one successful read so that
     * succeeding reads are preceeded by a select with a 0 timeout to prevent
     * read from blocking indefinitely.
     */
    XisbBlockDuration(pMse->buffer, -1);
    while((c = XisbRead(pMse->buffer)) >= 0) {
	u = (unsigned char)c;
	pBuf[pBufP++] = u;
	if (pBufP < sizeof(struct input_event))
	    continue;
	memcpy(&ev, pBuf, sizeof(struct input_event));
	if (ev.type == EV_SYN || ev.type == EV_MSC)
	    goto ignore_event;

	ev_type = LeventFindType(ev.type);
	if (!ev_type) {
	    if (mPriv->show_events)
		ErrorF("got event with unknown type: "
		       "type=%u code=%u value=%d\n",
		       ev.type, ev.code, ev.value);
	    goto ignore_event;
	}
	if (!ev_type->code_table) {
	    if (mPriv->show_events)
		ErrorF("got event w/o code table: "
		       "type=%s code=%u value=%d\n",
		       ev_type->name, ev.code, ev.value);
	    goto ignore_event;
	}
	ev_code = LeventFindCode(ev_type->code_table, ev.code);
	if (!ev_code) {
	    if (mPriv->show_events)
		ErrorF("got event with unknown code: "
		       "type=%s code=%u value=%d\n",
		       ev_type->name, ev.code, ev.value);
	    goto ignore_event;
	}

	dx = dy = 0;
	button = -1;
	if (mPriv->show_events)
	    ErrorF("got event: type=%s code=%s value=%d\n",
		   ev_type->name, ev_code->name, ev.value);
	switch(MAP_TYPE(ev_code->mapping)) {
	    
	  case MAPTYPE_NONE:
	    if (mPriv->show_events)
		ErrorF("not mapped\n");
	    break;

	  case MAPTYPE_X:
	    dx = ev.value;
	    if (mPriv->show_events)
		ErrorF("mapped to X axis\n");
	    break;
	
	  case MAPTYPE_Y:
	    dy = ev.value;
	    if (mPriv->show_events)
		ErrorF("mapped to Y axis\n");
	    break;

	  case MAPTYPE_BUTTON:
	    button = MAP_P0(ev_code->mapping);
	    down = !!ev.value;
	    if (mPriv->show_events)
		ErrorF("mapped to %s button %u\n",
		       down ? "press" : "release", button+1);
	    break;

	  case MAPTYPE_AXIS:
	    button = (ev.value < 0) ?
		     MAP_P0(ev_code->mapping) : MAP_P1(ev_code->mapping);
	    down = -1;
	    if (mPriv->show_events)
		ErrorF("mapped to press/release button %u\n", button+1);
	    break;
	}
	    
	if (dx || dy)
	    xf86PostMotionEvent(pInfo->dev, 0, 0, 2, dx, dy);
	if (button >= 0) {
	    if (down < 0) {
		xf86PostButtonEvent(pInfo->dev, 0, button+1, 1, 0, 0);
		xf86PostButtonEvent(pInfo->dev, 0, button+1, 0, 0, 0);
	    }
	    else {
		xf86PostButtonEvent(pInfo->dev, 0, button+1, down, 0, 0);
	    }
	}

      ignore_event:
	pBufP = 0;
    }
    pMse->protoBufTail = pBufP;
}


static void
LeventCtrl(DeviceIntPtr device, PtrCtrl *ctrl)
{
    InputInfoPtr pInfo;
    MouseDevPtr pMse;

    pInfo = device->public.devicePrivate;
    pMse = pInfo->private;

    pMse->num       = ctrl->num;
    pMse->den       = ctrl->den;
    pMse->threshold = ctrl->threshold;
}

static int
LeventProc(DeviceIntPtr device, int what)
{
    InputInfoPtr pInfo;
    MouseDevPtr pMse;
    mousePrivPtr mPriv;
    unsigned char map[MSE_MAXBUTTONS + 1];
    int i;
    
    pInfo = device->public.devicePrivate;
    pMse = pInfo->private;
    pMse->device = device;
    mPriv = (mousePrivPtr)pMse->mousePriv;    

    switch (what)
    {
    case DEVICE_INIT:
	device->public.on = FALSE;
	for (i = 0; i < MSE_MAXBUTTONS; i++)
	    map[i + 1] = i + 1;

	InitPointerDeviceStruct((DevicePtr)device, map,
				min(pMse->buttons, MSE_MAXBUTTONS),
				miPointerGetMotionEvents, pMse->Ctrl,
				miPointerGetMotionBufferSize());

	/* X valuator */
	xf86InitValuatorAxisStruct(device, 0, 0, -1, 1, 0, 1);
	xf86InitValuatorDefaults(device, 0);
	/* Y valuator */
	xf86InitValuatorAxisStruct(device, 1, 0, -1, 1, 0, 1);
	xf86InitValuatorDefaults(device, 1);
	xf86MotionHistoryAllocate(pInfo);
	break;

    case DEVICE_ON:
        /* open device */
	pInfo->fd = LeventOpenDevice(pInfo);
	if (pInfo->fd == -1)
	    xf86Msg(X_WARNING, "%s: cannot open input device\n", pInfo->name);
	else {
	    pMse->buffer = XisbNew(pInfo->fd, 64);
	    if (!pMse->buffer) {
		close(pInfo->fd);
		pInfo->fd = -1;
	    } else {
		mousePrivPtr mPriv = (mousePrivPtr)pMse->mousePriv;
		if (mPriv != NULL) {
		    xf86FlushInput(pInfo->fd);
		    xf86AddEnabledDevice(pInfo);
		}
	    }
	}
	pMse->lastButtons = 0;
	device->public.on = TRUE;
	FlushButtons(pMse);
	break;
	    
    case DEVICE_OFF:
    case DEVICE_CLOSE:
	if (pInfo->fd != -1) {
	    xf86RemoveEnabledDevice(pInfo);
	    if (pMse->buffer) {
		XisbFree(pMse->buffer);
		pMse->buffer = NULL;
	    }
	    close(pInfo->fd);
	    pInfo->fd = -1;
	}
	device->public.on = FALSE;
	usleep(300000);
	break;
    }
    return Success;
}

static Bool
LeventConvert(InputInfoPtr pInfo, int first, int num, int v0, int v1, int v2,
	      int v3, int v4, int v5, int *x, int *y)
{
    if (first != 0 || num != 2)
	return FALSE;

    *x = v0;
    *y = v1;

    return TRUE;
}

static void
FlushButtons(MouseDevPtr pMse)
{

    /* If no button down is pending xf86PostButtonEvent()
     * will discard them. So we are on the safe side. */

    int i, blocked;

    pMse->lastButtons = 0;

    blocked = xf86BlockSIGIO ();
    for (i = 1; i <= MSE_MAXBUTTONS; i++)
	xf86PostButtonEvent(pMse->device,0,i,0,0,0);
    xf86UnblockSIGIO (blocked);
}
